##
# Copyright 2020-2023 NVIDIA
#
# This file is triple-licensed under GPLv2 (see below), MIT, and
# BSD three-clause licenses.
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for XALT, implemented as an easyblock
@author: Scott McMillan (NVIDIA)
"""
import os

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import CUSTOM, MANDATORY
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_XALT(ConfigureMake):
    """Support for building and installing XALT."""

    @staticmethod
    def extra_options():
        extra_vars = {
            'config_py': [None, "XALT site filter file", MANDATORY],
            'executable_tracking': [True, "Enable executable tracking", CUSTOM],
            'gpu_tracking': [None, "Enable GPU tracking", CUSTOM],
            'logging_url': [None, "Logging URL for transmission", CUSTOM],
            'mysql': [False, "Build with MySQL support", CUSTOM],
            'scalar_sampling': [True, "Enable scalar sampling", CUSTOM],
            'static_cxx': [False, "Statically link libstdc++ and libgcc_s", CUSTOM],
            'syshost': [None, "System name", MANDATORY],
            'transmission': [None, "Data tranmission method", MANDATORY],
            'file_prefix': [None, "XALT record files prefix", CUSTOM],
        }
        return ConfigureMake.extra_options(extra_vars)

    def configure_step(self):
        """Custom configuration step for XALT."""

        # By default, XALT automatically appends 'xalt/<version>' to the
        # prefix, i.e., --prefix=/opt will actually install in
        # /opt/xalt/<version>.  To precisely control the install prefix and
        # not append anything to the prefix, use the configure option
        # '--with-siteControlledPrefix=yes'.
        # See https://xalt.readthedocs.io/en/latest/050_install_and_test.html
        self.cfg.update('configopts', '--with-siteControlledPrefix=yes')

        # XALT site filter config file is mandatory
        config_py = self.cfg['config_py']
        if config_py:
            if os.path.exists(config_py):
                self.cfg.update('configopts', '--with-config=%s' % config_py)
            else:
                raise EasyBuildError("Specified XALT configuration file %s does not exist!", config_py)
        else:
            error_msg = "Location of XALT configuration file must be specified via 'config_py' easyconfig parameter. "
            error_msg += "You can edit the easyconfig file, or use 'eb --try-amend=config_py=<path>'. "
            error_msg += "See https://xalt.readthedocs.io/en/latest/030_site_filtering.html for more information."
            raise EasyBuildError(error_msg)

        # XALT system name is mandatory
        if self.cfg['syshost']:
            self.cfg.update('configopts', '--with-syshostConfig=%s' % self.cfg['syshost'])
        else:
            error_msg = "The name of the system must be specified via the 'syshost' easyconfig parameter. "
            error_msg += "You can edit the easyconfig file, or use 'eb --try-amend=syshost=<string>'. "
            error_msg += "See https://xalt.readthedocs.io/en/latest/020_site_configuration.html for more information."
            raise EasyBuildError(error_msg)

        # Transmission method is mandatory
        if self.cfg['transmission']:
            self.cfg.update('configopts', '--with-transmission=%s' % self.cfg['transmission'])
        else:
            error_msg = "The XALT transmission method must be specified via the 'transmission' easyconfig parameter. "
            error_msg = "You can edit the easyconfig file, or use 'eb --try-amend=transmission=<string>'. "
            error_msg += "See https://xalt.readthedocs.io/en/latest/020_site_configuration.html for more information."
            raise EasyBuildError(error_msg)

        # GPU tracking
        if self.cfg['gpu_tracking'] is True:
            # User enabled
            self.cfg.update('configopts', '--with-trackGPU=yes')
        elif self.cfg['gpu_tracking'] is None:
            # Default value, enable GPU tracking if nvml.h is present
            # and the CUDA module is loaded
            cuda_root = get_software_root('CUDA')
            if cuda_root:
                nvml_h = os.path.join(cuda_root, "include", "nvml.h")
                if os.path.isfile(nvml_h):
                    self.cfg.update('configopts', '--with-trackGPU=yes')
                    self.cfg['gpu_tracking'] = True
        else:
            # User disabled
            self.cfg.update('configopts', '--with-trackGPU=no')

        # MySQL
        if self.cfg['mysql'] is True:
            self.cfg.update('configopts', '--with-MySQL=yes')
        else:
            self.cfg.update('configopts', '--with-MySQL=no')

        # If XALT is built with a more recent compiler than the system
        # compiler, then XALT likely will depend on symbol versions not
        # available in the system libraries. Link statically as a workaround.
        if self.cfg['static_cxx'] is True:
            self.cfg.update('configopts', 'LDFLAGS="${LDFLAGS} -static-libstdc++ -static-libgcc"')

        # XALT file prefix (optional). The default is $HOME/.xalt.d/ which
        # entails that record files are stored separately for each user.
        # If this option is specified, XALT will write to the specified
        # location for every user. The file prefix can also be modified
        # after the install using the XALT_FILE_PREFIX environment variable.
        if self.cfg['file_prefix']:
            self.cfg.update('configopts', '--with-xaltFilePrefix=%s' % self.cfg['file_prefix'])

        # Configure
        super(EB_XALT, self).configure_step()

    def make_module_extra(self, *args, **kwargs):
        txt = super(EB_XALT, self).make_module_extra(*args, **kwargs)

        txt += self.module_generator.prepend_paths('LD_PRELOAD', 'lib64/libxalt_init.%s' % get_shared_lib_ext())
        txt += self.module_generator.set_environment('XALT_DIR', self.installdir)
        txt += self.module_generator.set_environment('XALT_ETC_DIR', '%s' % os.path.join(self.installdir, 'etc'))
        txt += self.module_generator.set_environment('XALT_EXECUTABLE_TRACKING',
                                                     ('no', 'yes')[bool(self.cfg['executable_tracking'])])
        txt += self.module_generator.set_environment('XALT_GPU_TRACKING',
                                                     ('no', 'yes')[bool(self.cfg['gpu_tracking'])])
        if self.cfg['transmission'].lower() == 'curl' and self.cfg['logging_url']:
            txt += self.module_generator.set_environment('XALT_LOGGING_URL', self.cfg['logging_url'])
        txt += self.module_generator.set_environment('XALT_SCALAR_SAMPLING',
                                                     ('no', 'yes')[bool(self.cfg['scalar_sampling'])])

        # In order to track containerized executables, bind mount the XALT
        # directory in the Singularity container and preload the XALT library
        # https://xalt.readthedocs.io/en/latest/050_install_and_test.html#xalt-modulefile
        txt += self.module_generator.prepend_paths('SINGULARITY_BINDPATH', '')
        txt += self.module_generator.prepend_paths('SINGULARITYENV_LD_PRELOAD',
                                                   'lib64/libxalt_init.%s' % get_shared_lib_ext())

        return txt

    def make_module_req_guess(self):
        """Custom guesses for environment variables"""
        return {'COMPILER_PATH': 'bin',
                'PATH': 'bin'}

    def sanity_check_step(self):
        """Custom sanity check"""
        custom_paths = {
            'files': ['bin/ld', 'bin/ld.gold', 'bin/xalt_extract_record',
                      'lib64/libxalt_init.%s' % get_shared_lib_ext()],
            'dirs': ['bin', 'libexec', 'sbin'],
        }
        custom_commands = ['xalt_configuration_report']

        super(EB_XALT, self).sanity_check_step(custom_commands=custom_commands,
                                               custom_paths=custom_paths)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing reticulate R package, implemented as an easyblock

@author: Samuel Moors (Vrije Universiteit Brussel)
"""
import os

from easybuild.easyblocks.generic.rpackage import RPackage
from easybuild.tools.modules import get_software_root


class EB_reticulate(RPackage):
    """Support for installing the reticulate R package."""

    def run(self):
        """Add extra environment variables to modulefile"""

        txt = super(EB_reticulate, self).run()
        if not txt:
            txt = ""

        pythonroot = get_software_root('Python')
        if pythonroot:
            # make sure EB-provided Python is used, and that reticulate does not install it's own Python
            # see: https://rstudio.github.io/reticulate/reference/use_python.html
            # see: https://github.com/rstudio/reticulate/issues/894
            txt += self.module_generator.set_environment('RETICULATE_PYTHON', os.path.join(pythonroot, 'bin', 'python'))
        else:
            self.log.info("Python not included as dependency, so RETICULATE_PYTHON not set")

        self.log.info("adding to modulefile: %s" % txt)
        return txt



##
# Copyright 2017-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing cryptography, implemented as an easyblock

@author: Alexander Grund
"""
from distutils.version import LooseVersion

from easybuild.easyblocks.generic.pythonpackage import PythonPackage
from easybuild.tools.run import run_cmd


class EB_cryptography(PythonPackage):
    """Support for building/installing cryptography."""

    def __init__(self, *args, **kwargs):
        """Initialize cryptography easyblock."""
        super(EB_cryptography, self).__init__(*args, **kwargs)

        # cryptography compiles a library using pthreads but does not link against it
        # which causes 'undefined symbol: pthread_atfork'
        # see https://github.com/easybuilders/easybuild-easyconfigs/issues/9446
        # and upstream: https://github.com/pyca/cryptography/issues/5084
        self.cfg['prebuildopts'] += 'CFLAGS="$CFLAGS -pthread"'
        self.cfg['preinstallopts'] += 'CFLAGS="$CFLAGS -pthread"'
        self.log.info("Adding -pthread to prebuildopts & preinstallopts of cryptography.\n" +
                      "Final values: prebuildopts=%s and preinstallopts=%s",
                      self.cfg['prebuildopts'],
                      self.cfg['preinstallopts'])

    def sanity_check_step(self, *args, **kwargs):
        """Custom sanity check"""
        success, fail_msg = super(EB_cryptography, self).sanity_check_step(*args, **kwargs)
        if success:
            # Check module added in v0.7 leading to issue #9446 (see above)
            if LooseVersion(self.version) >= LooseVersion("0.7"):
                run_cmd("python -c 'from cryptography.hazmat.bindings.openssl import binding'")
        return success, fail_msg



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for BerkeleyGW, implemented as an easyblock

@author: Miguel Dias Costa (National University of Singapore)
"""
import os
from distutils.version import LooseVersion

import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.environment import setvar
from easybuild.tools.filetools import copy_file
from easybuild.tools.modules import get_software_root, get_software_version


class EB_BerkeleyGW(ConfigureMake):
    """Support for building and installing BerkeleyGW."""

    @staticmethod
    def extra_options():
        """Custom easyconfig parameters for BerkeleyGW."""
        extra_vars = {
            'with_scalapack': [True, "Enable ScaLAPACK support", CUSTOM],
            'unpacked': [False, "Use unpacked rather than packed representation of the Hamiltonian", CUSTOM],
        }
        return ConfigureMake.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Add extra config options specific to BerkeleyGW."""
        super(EB_BerkeleyGW, self).__init__(*args, **kwargs)

    def configure_step(self):
        """No configuration procedure for BerkeleyGW."""
        pass

    def build_step(self):
        """Custom build step for BerkeleyGW."""

        self.cfg['parallel'] = 1

        self.cfg['buildopts'] = 'all-flavors'

        copy_file(os.path.join('config', 'generic.mpi.linux.mk'), 'arch.mk')

        mpicc = os.environ['MPICC']
        mpicxx = os.environ['MPICXX']
        mpif90 = os.environ['MPIF90']

        paraflags = []
        var_suffix = ''
        if self.toolchain.options.get('openmp', None):
            paraflags.append('-DOMP')
            var_suffix = '_MT'
        if self.toolchain.options.get('usempi', None):
            paraflags.append('-DMPI')
            self.cfg.update('buildopts', 'C_PARAFLAG="-DPARA"')
        self.cfg.update('buildopts', 'PARAFLAG="%s"' % ' '.join(paraflags))

        if self.toolchain.options.get('debug', None):
            self.cfg.update('buildopts', 'DEBUGFLAG="-DDEBUG -DVERBOSE"')
        else:
            self.cfg.update('buildopts', 'DEBUGFLAG=""')

        self.cfg.update('buildopts', 'LINK="%s"' % mpif90)
        self.cfg.update('buildopts', 'C_LINK="%s"' % mpicxx)

        self.cfg.update('buildopts', 'FOPTS="%s"' % os.environ['FFLAGS'])
        self.cfg.update('buildopts', 'C_OPTS="%s"' % os.environ['CFLAGS'])

        self.cfg.update('buildopts', 'LAPACKLIB="%s"' % os.environ['LIBLAPACK' + var_suffix])
        self.cfg.update('buildopts', 'SCALAPACKLIB="%s"' % os.environ['LIBSCALAPACK' + var_suffix])

        mathflags = []
        if self.cfg['with_scalapack']:
            mathflags.append('-DUSESCALAPACK')
        if self.cfg['unpacked']:
            mathflags.append('-DUNPACKED')

        comp_fam = self.toolchain.comp_family()
        if comp_fam == toolchain.INTELCOMP:
            self.cfg.update('buildopts', 'COMPFLAG="-DINTEL"')
            self.cfg.update('buildopts', 'MOD_OPT="-module "')
            self.cfg.update('buildopts', 'F90free="%s -free"' % mpif90)
            self.cfg.update('buildopts', 'FCPP="cpp -C -P -ffreestanding"')
            self.cfg.update('buildopts', 'C_COMP="%s"' % mpicc)
            self.cfg.update('buildopts', 'CC_COMP="%s"' % mpicxx)
            self.cfg.update('buildopts', 'BLACSDIR="%s"' % os.environ['BLACS_LIB_DIR'])
            self.cfg.update('buildopts', 'BLACS="%s"' % os.environ['LIBBLACS'])
        elif comp_fam == toolchain.GCC:
            c_flags = "-std=c99"
            cxx_flags = "-std=c++0x"
            f90_flags = "-ffree-form -ffree-line-length-none -fno-second-underscore"
            if LooseVersion(get_software_version('GCC')) >= LooseVersion('10'):
                c_flags += " -fcommon"
                cxx_flags += " -fcommon"
                f90_flags += " -fallow-argument-mismatch"
            self.cfg.update('buildopts', 'COMPFLAG="-DGNU"')
            self.cfg.update('buildopts', 'MOD_OPT="-J "')
            self.cfg.update('buildopts', 'F90free="%s %s"' % (mpif90, f90_flags))
            self.cfg.update('buildopts', 'FCPP="cpp -C -nostdinc -nostdinc++"')
            self.cfg.update('buildopts', 'C_COMP="%s %s"' % (mpicc, c_flags))
            self.cfg.update('buildopts', 'CC_COMP="%s %s"' % (mpicxx, cxx_flags))
        else:
            raise EasyBuildError("EasyBuild does not yet have support for building BerkeleyGW with toolchain %s"
                                 % comp_fam)

        mkl = get_software_root('imkl')
        if mkl:
            self.cfg.update('buildopts', 'MKLPATH="%s"' % os.getenv('MKLROOT'))

        fftw = get_software_root('FFTW')
        if mkl or fftw:
            mathflags.append('-DUSEFFTW3')
            self.cfg.update('buildopts', 'FFTWINCLUDE="%s"' % os.environ['FFTW_INC_DIR'])

            libfft_var = 'LIBFFT%s' % var_suffix
            fft_libs = os.environ[libfft_var]

            if fftw and get_software_root('fftlib'):
                fft_libs = "%s %s" % (os.environ['FFTLIB'], fft_libs)

            self.cfg.update('buildopts', 'FFTWLIB="%s"' % fft_libs)

        hdf5 = get_software_root('HDF5')
        if hdf5:
            mathflags.append('-DHDF5')
            self.cfg.update('buildopts', 'HDF5INCLUDE="%s/include"' % hdf5)
            self.cfg.update('buildopts', 'HDF5LIB="-L%s/lib -lhdf5hl_fortran -lhdf5_hl -lhdf5_fortran -lhdf5 -lsz -lz"'
                            % hdf5)

        elpa = get_software_root('ELPA')
        if elpa:
            if not self.cfg['with_scalapack']:
                raise EasyBuildError("ELPA requires ScaLAPACK but 'with_scalapack' is set to False")
            mathflags.append('-DUSEELPA')
            elpa_suffix = '_openmp' if self.toolchain.options.get('openmp', None) else ''
            self.cfg.update('buildopts', 'ELPALIB="%s/lib/libelpa%s.a"' % (elpa, elpa_suffix))
            self.cfg.update('buildopts', 'ELPAINCLUDE="%s/include/elpa%s-%s/modules"'
                            % (elpa, elpa_suffix, get_software_version('ELPA')))

        self.cfg.update('buildopts', 'MATHFLAG="%s"' % ' '.join(mathflags))

        super(EB_BerkeleyGW, self).build_step()

    def install_step(self):
        """Custom install step for BerkeleyGW."""
        self.cfg.update('installopts', 'INSTDIR="%s"' % self.installdir)
        super(EB_BerkeleyGW, self).install_step()

    def test_step(self):
        """Custom test step for BerkeleyGW."""
        if self.cfg['runtest'] is not False:
            self.cfg['runtest'] = 'check'
            setvar('OMP_NUM_THREADS', '4')
            setvar('TEMPDIRPATH', os.path.join(self.builddir, 'tmp'))
        super(EB_BerkeleyGW, self).test_step()

    def sanity_check_step(self):
        """Custom sanity check for BerkeleyGW."""

        progs = ['epsilon', 'sigma', 'kernel', 'absorption', 'nonlinearoptics', 'parabands']
        flavors = ['real', 'cplx']
        files = [os.path.join('bin', prog + '.' + flavor + '.x') for prog in progs for flavor in flavors]

        custom_paths = {
            'files': files,
            'dirs': [],
        }

        super(EB_BerkeleyGW, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for CPLEX, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
"""
from distutils.version import LooseVersion
import glob
import os
import stat

import easybuild.tools.environment as env
from easybuild.easyblocks.generic.binary import Binary
from easybuild.easyblocks.generic.pythonpackage import det_pylibdir
from easybuild.easyblocks.python import EBPYTHONPREFIXES
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import adjust_permissions, change_dir, mkdir
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd, run_cmd_qa


class EB_CPLEX(Binary):
    """
    Support for installing CPLEX.
    Version 12.2 has a self-extracting binary with a Java installer
    """

    def __init__(self, *args, **kwargs):
        """Initialize CPLEX-specific variables."""
        super(EB_CPLEX, self).__init__(*args, **kwargs)

        self.bindir = None
        self.with_python = False
        self.multi_python = 'Python' in self.cfg['multi_deps']

    def prepare_step(self, *args, **kwargs):
        """Prepare build environment."""
        super(EB_CPLEX, self).prepare_step(*args, **kwargs)

        if get_software_root('Python'):
            self.with_python = True

    def install_step(self):
        """CPLEX has an interactive installer, so use Q&A"""

        tmpdir = os.path.join(self.builddir, 'tmp')
        stagedir = os.path.join(self.builddir, 'staged')
        change_dir(self.builddir)
        mkdir(tmpdir)
        mkdir(stagedir)

        env.setvar('IATEMPDIR', tmpdir)
        dst = os.path.join(self.builddir, self.src[0]['name'])

        cmd = "%s -i console" % dst

        qanda = {
            "PRESS <ENTER> TO CONTINUE:": '',
            'Press Enter to continue viewing the license agreement, or enter'
            ' "1" to accept the agreement, "2" to decline it, "3" to print it,'
            ' or "99" to go back to the previous screen.:': '1',
            'ENTER AN ABSOLUTE PATH, OR PRESS <ENTER> TO ACCEPT THE DEFAULT :': self.installdir,
            'IS THIS CORRECT? (Y/N):': 'y',
            'PRESS <ENTER> TO INSTALL:': '',
            "PRESS <ENTER> TO EXIT THE INSTALLER:": '',
            "CHOOSE LOCALE BY NUMBER:": '',
            "Choose Instance Management Option:": '',
            "No model content or proprietary data will be sent.\n1- Yes\n2- No\n"
            "ENTER THE NUMBER OF THE DESIRED CHOICE:": '2',
        }
        noqanda = [r'Installing\.\.\..*\n.*------.*\n\n.*============.*\n.*$']

        run_cmd_qa(cmd, qanda, no_qa=noqanda, log_all=True, simple=True)

        # fix permissions on install dir
        perms = stat.S_IRWXU | stat.S_IXOTH | stat.S_IXGRP | stat.S_IROTH | stat.S_IRGRP
        adjust_permissions(self.installdir, perms, recursive=False, relative=False)

        # also install Python bindings if Python is included as a dependency
        if self.with_python:
            cwd = change_dir(os.path.join(self.installdir, 'python'))
            run_cmd("python setup.py install --prefix=%s" % self.installdir)
            change_dir(cwd)

    def det_bindir(self):
        """Determine CPLEX bin subdirectory."""

        # avoid failing miserably under --module-only --force
        if os.path.exists(self.installdir) and os.listdir(self.installdir):
            bin_glob = 'cplex/bin/x86-64*'
            change_dir(self.installdir)
            bins = glob.glob(bin_glob)

            if len(bins) == 1:
                self.bindir = bins[0]
            elif len(bins) > 1:
                raise EasyBuildError("More than one possible path for bin found: %s", bins)
            else:
                raise EasyBuildError("No bins found using %s in %s", bin_glob, self.installdir)
        else:
            self.bindir = 'UNKNOWN'

    def make_module_extra(self):
        """Add bin dirs and lib dirs and set CPLEX_HOME and CPLEXDIR"""
        txt = super(EB_CPLEX, self).make_module_extra()

        # avoid failing miserably under --module-only --force
        if os.path.exists(self.installdir):
            cwd = change_dir(self.installdir)
            bins = glob.glob(os.path.join('*', 'bin', 'x86-64*'))
            libs = glob.glob(os.path.join('*', 'lib', 'x86-64*', '*pic'))
            change_dir(cwd)
        else:
            bins = []
            libs = []

        txt += self.module_generator.prepend_paths('PATH', [path for path in bins])
        txt += self.module_generator.prepend_paths('LD_LIBRARY_PATH', [path for path in bins + libs])

        txt += self.module_generator.set_environment('CPLEX_HOME', os.path.join(self.installdir, 'cplex'))
        txt += self.module_generator.set_environment('CPLEXDIR', os.path.join(self.installdir, 'cplex'))

        if self.with_python:
            if self.multi_python:
                txt += self.module_generator.prepend_paths(EBPYTHONPREFIXES, '')
            else:
                txt += self.module_generator.prepend_paths('PYTHONPATH', [det_pylibdir()])

        self.log.debug("make_module_extra added %s" % txt)
        return txt

    def sanity_check_step(self):
        """Custom sanity check for CPLEX"""

        if self.bindir is None:
            self.det_bindir()

        binaries = ['cplex', 'cplexamp']
        if LooseVersion(self.version) < LooseVersion('12.8'):
            binaries.append('convert')

        custom_paths = {
            'files': [os.path.join(self.bindir, x) for x in binaries],
            'dirs': [],
        }
        custom_commands = []

        if self.with_python:
            custom_commands.append("python -c 'import cplex'")
            custom_commands.append("python -c 'import docplex'")

        super(EB_CPLEX, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing R packages, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Toon Willems (Ghent University)
@author: Balazs Hajgato (Vrije Universiteit Brussel)
"""
import os
import re

from easybuild.easyblocks.r import EXTS_FILTER_R_PACKAGES, EB_R
from easybuild.easyblocks.generic.configuremake import check_config_guess, obtain_config_guess
from easybuild.framework.easyconfig import CUSTOM
from easybuild.framework.extensioneasyblock import ExtensionEasyBlock
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import mkdir, copy_file
from easybuild.tools.run import run_cmd, parse_log_for_error


def make_R_install_option(opt, values, cmdline=False):
    """
    Make option list for install.packages, to specify in R environment.
    """
    txt = ""
    if values:
        if cmdline:
            txt = " --%s=\"%s" % (opt, values[0])
        else:
            txt = "%s=c(\"%s" % (opt, values[0])
        for i in values[1:]:
            txt += " %s" % i
        if cmdline:
            txt += "\""
        else:
            txt += "\")"
    return txt


class RPackage(ExtensionEasyBlock):
    """
    Install an R package as a separate module, or as an extension.
    """

    @staticmethod
    def extra_options(extra_vars=None):
        """Extra easyconfig parameters specific to RPackage."""
        extra_vars = ExtensionEasyBlock.extra_options(extra_vars=extra_vars)
        extra_vars.update({
            'exts_subdir': ['', "Subdirectory where R extensions should be installed info", CUSTOM],
            'unpack_sources': [False, "Unpack sources before installation", CUSTOM],
        })
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Initliaze RPackage-specific class variables."""

        super(RPackage, self).__init__(*args, **kwargs)

        self.configurevars = []
        self.configureargs = []
        self.ext_src = None
        self._required_deps = None

    def make_r_cmd(self, prefix=None):
        """Create a command to run in R to install an R package."""
        confvars = "confvars"
        confargs = "confargs"
        confvarslist = make_R_install_option(confvars, self.configurevars)
        confargslist = make_R_install_option(confargs, self.configureargs)
        confvarsstr = ""
        if confvarslist:
            confvarslist = confvarslist + "; names(%s)=\"%s\"" % (confvars, self.name)
            confvarsstr = ", configure.vars=%s" % confvars
        confargsstr = ""
        if confargslist:
            confargslist = confargslist + "; names(%s)=\"%s\"" % (confargs, self.name)
            confargsstr = ", configure.args=%s" % confargs

        if prefix:
            prefix = '"%s", ' % prefix
        else:
            prefix = ''

        r_cmd = """
        options(repos=c(CRAN="http://www.freestatistics.org/cran"))
        %s
        %s
        install.packages("%s", %s dependencies = FALSE %s%s)
        """ % (confvarslist, confargslist, self.name, prefix, confvarsstr, confargsstr)
        cmd = "%s R -q --no-save %s" % (self.cfg['preinstallopts'], self.cfg['installopts'])

        self.log.debug("make_r_cmd returns %s with input %s" % (cmd, r_cmd))

        return (cmd, r_cmd)

    def make_cmdline_cmd(self, prefix=None):
        """Create a command line to install an R package."""
        confvars = ""
        if self.configurevars:
            confvars = make_R_install_option("configure-vars", self.configurevars, cmdline=True)
        confargs = ""
        if self.configureargs:
            confargs = make_R_install_option("configure-args", self.configureargs, cmdline=True)

        if prefix:
            prefix = '--library=%s' % prefix
        else:
            prefix = ''

        loc = self.start_dir
        if loc is None:
            loc = self.ext_dir or self.ext_src
        elif not os.path.isabs(loc):
            # TODO: deprecated behaviour in framework 4.7.1, remove after 5.0
            loc = os.path.join(self.ext_dir or os.path.sep, loc)
            deprecation_msg = "Found relative path in start_dir, please upgrade to easybuild-framework>=4.7.1"
            self.log.deprecated(deprecation_msg, '5.0')

        cmd = ' '.join([
            self.cfg['preinstallopts'],
            "R CMD INSTALL",
            loc,
            confargs,
            confvars,
            prefix,
            '--no-clean-on-error',
            self.cfg['installopts'],
        ])

        self.log.debug("make_cmdline_cmd returns %s" % cmd)
        return cmd, None

    def configure_step(self):
        """No configuration for installing R packages."""
        pass

    def build_step(self):
        """No separate build step for R packages."""
        pass

    def install_R_package(self, cmd, inp=None):
        """Install R package as specified, and check for errors."""

        output, _ = run_cmd(cmd, log_all=True, simple=False, inp=inp, regexp=False)
        self.check_install_output(output)

    def check_install_output(self, output):
        """
        Check output of installation command, and clean up installation if needed.
        """
        errors = parse_log_for_error(output, regExp="^ERROR:")
        if errors:
            self.handle_installation_errors()
            cmd = "R -q --no-save"
            stdin = """
            remove.library(%s)
            """ % self.name
            # remove package if errors were detected
            # it's possible that some of the dependencies failed, but the package itself was installed
            run_cmd(cmd, log_all=False, log_ok=False, simple=False, inp=stdin, regexp=False)
            raise EasyBuildError("Errors detected during installation of R package %s!", self.name)
        else:
            self.log.debug("R package %s installed succesfully", self.name)

    def update_config_guess(self, path):
        """Update any config.guess found in specified directory"""
        for config_guess_dir in (root for root, _, files in os.walk(path) if 'config.guess' in files):
            config_guess = os.path.join(config_guess_dir, 'config.guess')
            if not check_config_guess(config_guess):
                updated_config_guess = obtain_config_guess()
                if updated_config_guess:
                    self.log.debug("Replacing outdated %s with more recent %s", config_guess, updated_config_guess)
                    copy_file(updated_config_guess, config_guess)
                else:
                    raise EasyBuildError("Failed to obtain updated config.guess")

    def install_step(self):
        """Install procedure for R packages."""
        # Update config.guess if the package was extracted
        if self.start_dir:
            self.update_config_guess(self.start_dir)
        cmd, stdin = self.make_cmdline_cmd(prefix=os.path.join(self.installdir, self.cfg['exts_subdir']))
        self.install_R_package(cmd, inp=stdin)

    @property
    def required_deps(self):
        """Return list of required dependencies for this extension."""

        if self._required_deps is None:
            if self.src:
                cmd = "tar --wildcards --extract --file %s --to-stdout '*/DESCRIPTION'" % self.src
                out, _ = run_cmd(cmd, simple=False, trace=False)

                # lines that start with whitespace are merged with line above
                lines = []
                for line in out.splitlines():
                    if line and line[0] in (' ', '\t'):
                        lines[-1] = lines[-1] + line
                    else:
                        lines.append(line)
                out = '\n'.join(lines)

                pkg_key = 'Package:'
                deps_map = {}
                deps = []
                pkg = None

                for line in out.splitlines():
                    if pkg_key in line:
                        if pkg is not None:
                            deps = []

                        pkg_name_regex = re.compile(r'Package:\s*([^ ]+)')
                        res = pkg_name_regex.search(line)
                        if res:
                            pkg = res.group(1)
                            if pkg in deps_map:
                                deps = deps_map[pkg]
                        else:
                            raise EasyBuildError("Failed to determine package name from line '%s'", line)

                        deps_map[pkg] = deps

                    elif any(line.startswith(x) for x in ('Depends:', 'Imports:', 'LinkingTo:')):
                        # entries may specify version requirements between brackets (which we don't care about here)
                        dep_names = [x.split('(')[0].strip() for x in line.split(':', 1)[1].split(',')]
                        deps.extend([d for d in dep_names if d not in ('', 'R', self.name)])

                self._required_deps = deps_map.get(self.name, [])
                self.log.info("Required dependencies for %s: %s", self.name, self._required_deps)
            else:
                # no source => no required dependencies assumed
                self._required_deps = []

        return self._required_deps

    def prepare_r_ext_install(self):
        """
        Prepare installation of R package as extension.

        :return: Shell command to run + string to pass to stdin.
        """

        # determine location
        if isinstance(self.master, EB_R):
            # extension is being installed as part of an R installation/module
            (out, _) = run_cmd("R RHOME", log_all=True, simple=False, trace=False)
            rhome = out.strip()
            lib_install_prefix = os.path.join(rhome, 'library')
        else:
            # extension is being installed in a separate installation prefix
            lib_install_prefix = os.path.join(self.installdir, self.cfg['exts_subdir'])
            mkdir(lib_install_prefix, parents=True)

        if self.src:
            super(RPackage, self).run(unpack_src=True)
            self.ext_src = self.src
            self.update_config_guess(self.ext_dir)
            self.log.debug("Installing R package %s version %s." % (self.name, self.version))
            cmd, stdin = self.make_cmdline_cmd(prefix=lib_install_prefix)
        else:
            if self.patches:
                raise EasyBuildError("Cannot patch R package %s as no explicit source is given!", self.name)
            self.log.debug("Installing most recent version of R package %s (source not found)." % self.name)
            cmd, stdin = self.make_r_cmd(prefix=lib_install_prefix)

        return cmd, stdin

    def run(self):
        """
        Install R package as an extension.
        """
        cmd, stdin = self.prepare_r_ext_install()
        self.install_R_package(cmd, inp=stdin)

    def run_async(self):
        """
        Start installation of R package as an extension asynchronously.
        """
        cmd, stdin = self.prepare_r_ext_install()
        self.async_cmd_start(cmd, inp=stdin)

    def async_cmd_check(self):
        """
        Check progress of installation command that was started asynchronously.

        Output is checked for errors on completion.

        :return: True if command completed, False otherwise
        """
        done = super(RPackage, self).async_cmd_check()
        if done:
            self.check_install_output(self.async_cmd_output)

        return done

    def sanity_check_step(self, *args, **kwargs):
        """
        Custom sanity check for R packages
        """
        return super(RPackage, self).sanity_check_step(EXTS_FILTER_R_PACKAGES, *args, **kwargs)

    def make_module_extra(self):
        """Add install path to R_LIBS_SITE"""
        # prepend R_LIBS_SITE with install path
        extra = self.module_generator.prepend_paths("R_LIBS_SITE", [self.cfg['exts_subdir']])
        return super(RPackage, self).make_module_extra(extra)



##
# Copyright 2019-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for OpenMPI, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Robert Mijakovic (LuxProvide)
"""
import os
import re
from distutils.version import LooseVersion

import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig.constants import EASYCONFIG_CONSTANTS
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.config import build_option
from easybuild.tools.modules import get_software_root
from easybuild.tools.systemtools import check_os_dependency, get_shared_lib_ext
from easybuild.tools.toolchain.mpi import get_mpi_cmd_template


class EB_OpenMPI(ConfigureMake):
    """OpenMPI easyblock."""

    def configure_step(self):
        """Custom configuration step for OpenMPI."""

        def config_opt_used(key, enable_opt=False):
            """Helper function to check whether a configure option is already specified in 'configopts'."""
            if enable_opt:
                regex = '--(disable|enable)-%s' % key
            else:
                regex = '--(with|without)-%s' % key

            return bool(re.search(regex, self.cfg['configopts']))

        config_opt_names = [
            # suppress failure modes in relation to mpirun path
            'mpirun-prefix-by-default',
            # build shared libraries
            'shared',
        ]

        for key in config_opt_names:
            if not config_opt_used(key, enable_opt=True):
                self.cfg.update('configopts', '--enable-%s' % key)

        # List of EasyBuild dependencies for which OMPI has known options
        known_dependencies = ('CUDA', 'hwloc', 'libevent', 'libfabric', 'PMIx', 'UCX', 'UCC')
        # Value to use for `--with-<dep>=<value>` if the dependency is not specified in the easyconfig
        # No entry is interpreted as no option added at all
        # This is to make builds reproducible even when the system libraries are changed and avoids failures
        # due to e.g. finding only PMIx but not libevent on the system
        unused_dep_value = dict()
        # Known options since version 3.0 (no earlier ones checked)
        if LooseVersion(self.version) >= LooseVersion('3.0'):
            # Default to disable the option with "no"
            unused_dep_value = {dep: 'no' for dep in known_dependencies}
            # For these the default is to use an internal copy and not using any is not supported
            for dep in ('hwloc', 'libevent', 'PMIx'):
                unused_dep_value[dep] = 'internal'

        # handle dependencies
        for dep in known_dependencies:
            opt_name = dep.lower()
            # If the option is already used, don't add it
            if config_opt_used(opt_name):
                continue

            # libfabric option renamed in OpenMPI 3.1.0 to ofi
            if dep == 'libfabric' and LooseVersion(self.version) >= LooseVersion('3.1'):
                opt_name = 'ofi'
                # Check new option name. They are synonyms since 3.1.0 for backward compatibility
                if config_opt_used(opt_name):
                    continue

            dep_root = get_software_root(dep)
            # If the dependency is loaded, specify its path, else use the "unused" value, if any
            if dep_root:
                opt_value = dep_root
            else:
                opt_value = unused_dep_value.get(dep)
            if opt_value is not None:
                self.cfg.update('configopts', '--with-%s=%s' % (opt_name, opt_value))

        if bool(get_software_root('PMIx')) != bool(get_software_root('libevent')):
            raise EasyBuildError('You must either use both PMIx and libevent as dependencies or none of them. '
                                 'This is to enforce the same libevent is used for OpenMPI as for PMIx or '
                                 'the behavior may be unpredictable.')

        # check whether VERBS support should be enabled
        if not config_opt_used('verbs'):

            # for OpenMPI v4.x, the openib BTL should be disabled when UCX is used;
            # this is required to avoid "error initializing an OpenFabrics device" warnings,
            # see also https://www.open-mpi.org/faq/?category=all#ofa-device-error
            is_ucx_enabled = ('--with-ucx' in self.cfg['configopts'] and
                              '--with-ucx=no' not in self.cfg['configopts'])
            if LooseVersion(self.version) >= LooseVersion('4.0.0') and is_ucx_enabled:
                verbs = False
            else:
                # auto-detect based on available OS packages
                os_packages = EASYCONFIG_CONSTANTS['OS_PKG_IBVERBS_DEV'][0]
                verbs = any(check_os_dependency(osdep) for osdep in os_packages)
            # for OpenMPI v5.x, the verbs support is removed, only UCX is available
            # see https://github.com/open-mpi/ompi/pull/6270
            if LooseVersion(self.version) <= LooseVersion('5.0.0'):
                if verbs:
                    self.cfg.update('configopts', '--with-verbs')
                else:
                    self.cfg.update('configopts', '--without-verbs')

        super(EB_OpenMPI, self).configure_step()

    def test_step(self):
        """Test step for OpenMPI"""
        # Default to `make check` if nothing is set. Disable with "runtest = False" in the EC
        if self.cfg['runtest'] is None:
            self.cfg['runtest'] = 'check'

        super(EB_OpenMPI, self).test_step()

    def load_module(self, *args, **kwargs):
        """
        Load (temporary) module file, after resetting to initial environment.

        Also put RPATH wrappers back in place if needed, to ensure that sanity check commands work as expected.
        """
        super(EB_OpenMPI, self).load_module(*args, **kwargs)

        # ensure RPATH wrappers are in place, otherwise compiling minimal test programs will fail
        if build_option('rpath'):
            if self.toolchain.options.get('rpath', True):
                self.toolchain.prepare_rpath_wrappers(rpath_filter_dirs=self.rpath_filter_dirs,
                                                      rpath_include_dirs=self.rpath_include_dirs)

    def sanity_check_step(self):
        """Custom sanity check for OpenMPI."""

        bin_names = ['mpicc', 'mpicxx', 'mpif90', 'mpifort', 'mpirun', 'ompi_info', 'opal_wrapper']
        if LooseVersion(self.version) >= LooseVersion('5.0.0'):
            bin_names.append('prterun')
        else:
            bin_names.append('orterun')
        bin_files = [os.path.join('bin', x) for x in bin_names]

        shlib_ext = get_shared_lib_ext()
        lib_names = ['mpi_mpifh', 'mpi', 'open-pal']
        if LooseVersion(self.version) >= LooseVersion('5.0.0'):
            lib_names.append('prrte')
        else:
            lib_names.extend(['ompitrace', 'open-rte'])
        lib_files = [os.path.join('lib', 'lib%s.%s' % (x, shlib_ext)) for x in lib_names]

        inc_names = ['mpi-ext', 'mpif-config', 'mpif', 'mpi', 'mpi_portable_platform']
        if LooseVersion(self.version) >= LooseVersion('5.0.0'):
            inc_names.append('prte')
        inc_files = [os.path.join('include', x + '.h') for x in inc_names]

        custom_paths = {
            'files': bin_files + inc_files + lib_files,
            'dirs': [],
        }

        # make sure MPI compiler wrappers pick up correct compilers
        expected = {
            'mpicc': os.getenv('CC', 'gcc'),
            'mpicxx': os.getenv('CXX', 'g++'),
            'mpifort': os.getenv('FC', 'gfortran'),
            'mpif90': os.getenv('F90', 'gfortran'),
        }
        # actual pattern for gfortran is "GNU Fortran"
        for key in ['mpifort', 'mpif90']:
            if expected[key] == 'gfortran':
                expected[key] = "GNU Fortran"
        # for PGI, correct pattern is "pgfortran" with mpif90
        if expected['mpif90'] == 'pgf90':
            expected['mpif90'] = 'pgfortran'

        custom_commands = ["%s --version | grep '%s'" % (key, expected[key]) for key in sorted(expected.keys())]

        # Add minimal test program to sanity checks
        # Run with correct MPI launcher
        mpi_cmd_tmpl, params = get_mpi_cmd_template(toolchain.OPENMPI, dict(), mpi_version=self.version)
        # Limit number of ranks to 8 to avoid it failing due to hyperthreading
        ranks = min(8, self.cfg['parallel'])
        for src, compiler in (('hello_c.c', 'mpicc'), ('hello_mpifh.f', 'mpifort'), ('hello_usempi.f90', 'mpif90')):
            src_path = os.path.join(self.cfg['start_dir'], 'examples', src)
            if os.path.exists(src_path):
                test_exe = os.path.join(self.builddir, 'mpi_test_' + os.path.splitext(src)[0])
                self.log.info("Adding minimal MPI test program to sanity checks: %s", test_exe)

                # Build test binary
                custom_commands.append("%s %s -o %s" % (compiler, src_path, test_exe))

                # Run the test if chosen
                if build_option('mpi_tests'):
                    params.update({'nr_ranks': ranks, 'cmd': test_exe})
                    # Allow oversubscription for this test (in case of hyperthreading)
                    custom_commands.append("OMPI_MCA_rmaps_base_oversubscribe=1 " + mpi_cmd_tmpl % params)
                    # Run with 1 process which may trigger other bugs
                    # See https://github.com/easybuilders/easybuild-easyconfigs/issues/12978
                    params['nr_ranks'] = 1
                    custom_commands.append(mpi_cmd_tmpl % params)

        super(EB_OpenMPI, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing Pasha, implemented as an easyblock

@author: Jens Timmerman (Ghent University)
"""

import shutil
import os

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root


class EB_Pasha(ConfigureMake):
    """Support for building and installing Pasha"""

    def configure_step(self):
        """Configure Pasha by setting make options."""

        tbb = get_software_root('TBB')
        if not tbb:
            raise EasyBuildError("TBB module not loaded.")

        self.cfg.update('buildopts', "TBB_DIR=%s/tbb MPI_DIR='' MPI_INC='' " % tbb)
        self.cfg.update('buildopts', 'MPI_CXX="%s"' % os.getenv('MPICXX'))
        self.cfg.update('buildopts', 'OPM_FLAG="%s"' % self.toolchain.get_flag('openmp'))
        self.cfg.update('buildopts', 'MPI_LIB="" MY_CXX="%s" MPICH_IGNORE_CXX_SEEK=1' % os.getenv('CXX'))

    def install_step(self):
        """Install by copying everything from 'bin' subdir in build dir to install dir"""

        srcdir = os.path.join(self.builddir, "%s-%s" % (self.name, self.version), 'bin')
        shutil.copytree(srcdir, os.path.join(self.installdir, 'bin'))

    def sanity_check_step(self):
        """Custom sanity check for Pasha"""

        custom_paths = {
            'files': ["bin/pasha-%s" % x for x in ["kmergen", "pregraph", "graph"]],
            'dirs': [],
        }

        super(EB_Pasha, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2022-2023 Vrije Universiteit Brussel
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for bundles of Julia packages, implemented as an easyblock

@author: Alex Domingo (Vrije Universiteit Brussel)
"""
import os

from easybuild.easyblocks.generic.bundle import Bundle
from easybuild.easyblocks.generic.juliapackage import EXTS_FILTER_JULIA_PACKAGES, JuliaPackage
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root


class JuliaBundle(Bundle):
    """
    Bundle of JuliaPackages: install Julia packages as extensions in a bundle
    Defines custom sanity checks and module environment
    """

    @staticmethod
    def extra_options(extra_vars=None):
        """Easyconfig parameters specific to bundles of Julia packages."""
        if extra_vars is None:
            extra_vars = {}
        # combine custom easyconfig parameters of Bundle & JuliaPackage
        extra_vars = Bundle.extra_options(extra_vars)
        return JuliaPackage.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Initialize JuliaBundle easyblock."""
        super(JuliaBundle, self).__init__(*args, **kwargs)

        self.cfg['exts_defaultclass'] = 'JuliaPackage'
        self.cfg['exts_filter'] = EXTS_FILTER_JULIA_PACKAGES

        # need to disable templating to ensure that actual value for exts_default_options is updated...
        with self.cfg.disable_templating():
            # set default options for extensions according to relevant top-level easyconfig parameters
            jlpkg_keys = JuliaPackage.extra_options().keys()
            for key in jlpkg_keys:
                if key not in self.cfg['exts_default_options']:
                    self.cfg['exts_default_options'][key] = self.cfg[key]

            # Sources of Julia packages are commonly distributed from GitHub repos.
            # By default, rename downloaded tarballs to avoid name collisions on
            # packages sharing the same version string
            if 'sources' not in self.cfg['exts_default_options']:
                self.cfg['exts_default_options']['sources'] = [
                    {
                        'download_filename': 'v%(version)s.tar.gz',
                        'filename': '%(name)s-%(version)s.tar.gz',
                    }
                ]

        self.log.info("exts_default_options: %s", self.cfg['exts_default_options'])

    def prepare_step(self, *args, **kwargs):
        """Prepare for installing bundle of Julia packages."""
        super(JuliaBundle, self).prepare_step(*args, **kwargs)

        if get_software_root('Julia') is None:
            raise EasyBuildError("Julia not included as dependency!")

    def make_module_extra(self, *args, **kwargs):
        """Prepend installation directory to JULIA_DEPOT_PATH in module file."""
        txt = super(JuliaBundle, self).make_module_extra()
        txt += self.module_generator.prepend_paths('JULIA_DEPOT_PATH', [''])
        return txt

    def sanity_check_step(self, *args, **kwargs):
        """Custom sanity check for bundle of Julia packages"""
        custom_paths = {
            'files': [],
            'dirs': [os.path.join('packages', self.name)],
        }
        super(JuliaBundle, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing MrBayes, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Andy Georges (Ghent University)
@author: Maxime Boissonneault (Compute Canada, Calcul Quebec, Universite Laval)
@author: Jasper Grimm (University of York)
"""

import os
from distutils.version import LooseVersion

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import copy_file, mkdir
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd


class EB_MrBayes(ConfigureMake):
    """Support for building/installing MrBayes."""

    def __init__(self, *args, **kwargs):
        super(EB_MrBayes, self).__init__(*args, **kwargs)
        # For later MrBayes versions, no longer need to use this easyblock
        last_supported_version = '3.2.6'
        if LooseVersion(self.version) > LooseVersion(last_supported_version):
            raise EasyBuildError("Please use the ConfigureMake easyblock for %s versions > %s", self.name,
                                 last_supported_version)

    def configure_step(self):
        """Configure build: <single-line description how this deviates from standard configure>"""

        # set generic make options
        self.cfg.update('buildopts', 'CC="%s" OPTFLAGS="%s"' % (os.getenv('MPICC'), os.getenv('CFLAGS')))

        if LooseVersion(self.version) >= LooseVersion("3.2"):

            # set correct start_dir dir, and change into it
            # test whether it already contains 'src', since a reprod easyconfig would
            if os.path.basename(self.cfg['start_dir']) != 'src':
                self.cfg['start_dir'] = os.path.join(self.cfg['start_dir'], 'src')
            try:
                os.chdir(self.cfg['start_dir'])
            except OSError as err:
                raise EasyBuildError("Failed to change to correct source dir %s: %s", self.cfg['start_dir'], err)

            # run autoconf to generate configure script
            cmd = "autoconf"
            run_cmd(cmd)

            # set config opts
            beagle = get_software_root('beagle-lib')
            if beagle:
                self.cfg.update('configopts', '--with-beagle=%s' % beagle)
            else:
                if get_software_root('BEAGLE'):
                    self.log.nosupport('BEAGLE module as dependency, should be beagle-lib', '2.0')
                raise EasyBuildError("beagle-lib module not loaded?")

            if self.toolchain.options.get('usempi', None):
                self.cfg.update('configopts', '--enable-mpi')

            # configure
            super(EB_MrBayes, self).configure_step()
        else:

            # no configure script prior to v3.2
            self.cfg.update('buildopts', 'MPI=yes')

    def install_step(self):
        """Install by copying bniaries to install dir."""

        bindir = os.path.join(self.installdir, 'bin')
        mkdir(bindir)

        src = os.path.join(self.cfg['start_dir'], 'mb')
        dst = os.path.join(bindir, 'mb')
        copy_file(src, dst)
        self.log.info("Successfully copied %s to %s", src, dst)

    def sanity_check_step(self):
        """Custom sanity check for MrBayes."""

        custom_paths = {
            'files': ["bin/mb"],
            'dirs': [],
        }

        custom_commands = ["mb <<< %s" % x for x in ["about", "help"]]

        super(EB_MrBayes, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing Python, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Bart Oldeman (McGill University, Calcul Quebec, Compute Canada)
"""
import glob
import os
import re
import fileinput
import sys
import tempfile
from distutils.version import LooseVersion

import easybuild.tools.environment as env
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.framework.easyconfig.templates import TEMPLATE_CONSTANTS
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.config import build_option, log_path
from easybuild.tools.modules import get_software_libdir, get_software_root, get_software_version
from easybuild.tools.filetools import apply_regex_substitutions, change_dir, mkdir
from easybuild.tools.filetools import read_file, remove_dir, symlink, write_file
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext
import easybuild.tools.toolchain as toolchain


EXTS_FILTER_PYTHON_PACKAGES = ('python -c "import %(ext_name)s"', "")

# magic value for unlimited stack size
UNLIMITED = 'unlimited'

EBPYTHONPREFIXES = 'EBPYTHONPREFIXES'

# We want the following import order:
# 1. Packages installed into VirtualEnv
# 2. Packages installed into $EBPYTHONPREFIXES (e.g. our modules)
# 3. Packages installed in the Python module
# Note that this script is run after all sys.path manipulation by Python and Virtualenv are done.
# Hence prepending $EBPYTHONPREFIXES would shadow VirtualEnv packages and
# appending would NOT shadow the Python-module packages which makes updating packages via ECs impossible.
# Hence we move all paths which are prefixed with the Python-module path to the back but need to make sure
# not to move the VirtualEnv paths.
SITECUSTOMIZE = """
# sitecustomize.py script installed by EasyBuild,
# to pick up Python packages installed with `--prefix` into folders listed in $%(EBPYTHONPREFIXES)s

import os
import site
import sys

# print debug messages when $%(EBPYTHONPREFIXES)s_DEBUG is defined
debug = os.getenv('%(EBPYTHONPREFIXES)s_DEBUG')

# use prefixes from $EBPYTHONPREFIXES, so they have lower priority than
# virtualenv-installed packages, unlike $PYTHONPATH

ebpythonprefixes = os.getenv('%(EBPYTHONPREFIXES)s')

if ebpythonprefixes:
    postfix = os.path.join('lib', 'python' + '.'.join(map(str, sys.version_info[:2])), 'site-packages')
    if debug:
        print("[%(EBPYTHONPREFIXES)s] postfix subdirectory to consider in installation directories: %%s" %% postfix)

    potential_sys_prefixes = (getattr(sys, attr, None) for attr in ("real_prefix", "base_prefix", "prefix"))
    sys_prefix = next(p for p in potential_sys_prefixes if p)
    base_paths = [p for p in sys.path if p.startswith(sys_prefix)]

    for prefix in ebpythonprefixes.split(os.pathsep):
        if debug:
            print("[%(EBPYTHONPREFIXES)s] prefix: %%s" %% prefix)
        sitedir = os.path.join(prefix, postfix)
        if os.path.isdir(sitedir):
            if debug:
                print("[%(EBPYTHONPREFIXES)s] adding site dir: %%s" %% sitedir)
            site.addsitedir(sitedir)

    # Move base python paths to the end of sys.path so modules can override packages from the core Python module
    sys.path = [p for p in sys.path if p not in base_paths] + base_paths
""" % {'EBPYTHONPREFIXES': EBPYTHONPREFIXES}


class EB_Python(ConfigureMake):
    """Support for building/installing Python
    - default configure/build_step/make install works fine

    To extend Python by adding extra packages there are two ways:
    - list the packages in the exts_list, this will include the packages in this Python installation
    - create a seperate easyblock, so the packages can be loaded with module load

    e.g., you can include numpy and scipy in a default Python installation
    but also provide newer updated numpy and scipy versions by creating a PythonPackage-derived easyblock for it.
    """

    @staticmethod
    def extra_options():
        """Add extra config options specific to Python."""
        extra_vars = {
            'ebpythonprefixes': [True, "Create sitecustomize.py and allow use of $EBPYTHONPREFIXES", CUSTOM],
            'install_pip': [False,
                            "Use the ensurepip module (Python 2.7.9+, 3.4+) to install the bundled versions "
                            "of pip and setuptools into Python. You _must_ then use pip for upgrading "
                            "pip & setuptools by installing newer versions as extensions!",
                            CUSTOM],
            'optimized': [True, "Build with expensive, stable optimizations (PGO, etc.) (version >= 3.5.4)", CUSTOM],
            'ulimit_unlimited': [False, "Ensure stack size limit is set to '%s' during build" % UNLIMITED, CUSTOM],
            'use_lto': [None, "Build with Link Time Optimization (>= v3.7.0, potentially unstable on some toolchains). "
                        "If None: auto-detect based on toolchain compiler (version)", CUSTOM],
        }
        return ConfigureMake.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Constructor for Python easyblock."""
        super(EB_Python, self).__init__(*args, **kwargs)

        self.pyshortver = '.'.join(self.version.split('.')[:2])

        # Used for EBPYTHONPREFIXES handler script
        self.pythonpath = os.path.join(log_path(), 'python')

        ext_defaults = {
            # Use PYPI_SOURCE as the default for source_urls of extensions.
            'source_urls': [url for name, url, _ in TEMPLATE_CONSTANTS if name == 'PYPI_SOURCE'],
            # We should enable this (by default) for all extensions because the only installed packages at this point
            # (i.e. those in the site-packages folder) are the default installed ones, e.g. pip & setuptools.
            # And we must upgrade them cleanly, i.e. uninstall them first. This also applies to any other package
            # which is voluntarily or accidentally installed multiple times.
            # Example: Upgrading to a higher version after installing new dependencies.
            'pip_ignore_installed': False,
            # Python installations must be clean. Requires pip >= 9
            'sanity_pip_check': LooseVersion(self._get_pip_ext_version() or '0.0') >= LooseVersion('9.0'),
        }

        exts_default_options = self.cfg.get_ref('exts_default_options')
        for key in ext_defaults:
            if key not in exts_default_options:
                exts_default_options[key] = ext_defaults[key]
        self.log.debug("exts_default_options: %s", self.cfg['exts_default_options'])

        self.install_pip = self.cfg['install_pip']
        if self.install_pip:
            if not self._has_ensure_pip():
                raise EasyBuildError("The ensurepip module required to install pip (requested by install_pip=True) "
                                     "is not available in Python %s", self.version)

    def _get_pip_ext_version(self):
        """Return the pip version from exts_list or None"""
        for ext in self.cfg.get_ref('exts_list'):
            # Must be (at least) a name-version tuple
            if isinstance(ext, tuple) and len(ext) >= 2 and ext[0] == 'pip':
                return ext[1]
        return None

    def patch_step(self, *args, **kwargs):
        """
        Custom patch step for Python:
        * patch setup.py when --sysroot EasyBuild configuration setting is used
        """

        super(EB_Python, self).patch_step(*args, **kwargs)

        if self.install_pip:
            # Ignore user site dir. -E ignores PYTHONNOUSERSITE, so we have to add -s
            apply_regex_substitutions('configure', [(r"(PYTHON_FOR_BUILD=.*-E)'", r"\1 -s'")])

        # if we're installing Python with an alternate sysroot,
        # we need to patch setup.py which includes hardcoded paths like /usr/include and /lib64;
        # this fixes problems like not being able to build the _ssl module ("Could not build the ssl module")
        sysroot = build_option('sysroot')
        if sysroot:
            sysroot_inc_dirs, sysroot_lib_dirs = [], []

            for pattern in ['include*', os.path.join('usr', 'include*')]:
                sysroot_inc_dirs.extend(glob.glob(os.path.join(sysroot, pattern)))

            if sysroot_inc_dirs:
                sysroot_inc_dirs = ', '.join(["'%s'" % x for x in sysroot_inc_dirs])
            else:
                raise EasyBuildError("No include directories found in sysroot %s!", sysroot)

            for pattern in ['lib*', os.path.join('usr', 'lib*')]:
                sysroot_lib_dirs.extend(glob.glob(os.path.join(sysroot, pattern)))

            if sysroot_lib_dirs:
                sysroot_lib_dirs = ', '.join(["'%s'" % x for x in sysroot_lib_dirs])
            else:
                raise EasyBuildError("No lib directories found in sysroot %s!", sysroot)

            setup_py_fn = 'setup.py'
            setup_py_txt = read_file(setup_py_fn)

            # newer Python versions (3.6+) have refactored code, requires different patching approach
            if "system_include_dirs = " in setup_py_txt:
                regex_subs = [
                    (r"(system_include_dirs = \[).*\]", r"\1%s]" % sysroot_inc_dirs),
                    (r"(system_lib_dirs = \[).*\]", r"\1%s]" % sysroot_lib_dirs),
                ]
            else:
                regex_subs = [
                    (r"^([ ]+)'/usr/include',", r"\1%s," % sysroot_inc_dirs),
                    (r"\['/usr/include'\]", r"[%s]" % sysroot_inc_dirs),
                    (r"^([ ]+)'/lib64', '/usr/lib64',", r"\1%s," % sysroot_lib_dirs),
                    (r"^[ ]+'/lib', '/usr/lib',", ''),
                ]

            # Replace remaining hardcoded paths like '/usr/include', '/usr/lib' or '/usr/local',
            # where these paths are appearing inside single quotes (').
            # Inject sysroot in front to avoid picking up anything outside of sysroot,
            # We can leverage the single quotes such that we do not accidentally fiddle with other entries,
            # like /prefix/usr/include .
            for usr_subdir in ('usr/include', 'usr/lib', 'usr/local'):
                sysroot_usr_subdir = os.path.join(sysroot, usr_subdir)
                regex_subs.append((r"'/%s" % usr_subdir, r"'%s" % sysroot_usr_subdir))
                regex_subs.append((r'"/%s' % usr_subdir, r'"%s' % sysroot_usr_subdir))

            apply_regex_substitutions(setup_py_fn, regex_subs)

    def prepare_for_extensions(self):
        """
        Set default class and filter for Python packages
        """
        # build and install additional packages with PythonPackage easyblock
        self.cfg['exts_defaultclass'] = "PythonPackage"
        self.cfg['exts_filter'] = EXTS_FILTER_PYTHON_PACKAGES

        # don't add user site directory to sys.path (equivalent to python -s)
        env.setvar('PYTHONNOUSERSITE', '1')

        # don't pass down any build/install options that may have been specified
        # 'make' options do not make sense for when building/installing Python libraries (usually via 'python setup.py')
        msg = "Unsetting '%s' easyconfig parameter before building/installing extensions: %s"
        for param in ['buildopts', 'installopts']:
            if self.cfg[param]:
                self.log.debug(msg, param, self.cfg[param])
            self.cfg[param] = ''

        if self.install_pip:
            # When using ensurepip, then pip must be used to upgrade pip and setuptools
            # Otherwise it will only copy new files leading to a combination of files from the old and new version
            use_pip_default = self.cfg['exts_default_options'].get('use_pip')
            # self.exts is populated in fetch_step
            for ext in self.exts:
                if ext['name'] in ('pip', 'setuptools'):
                    if not ext.get('options', {}).get('use_pip', use_pip_default):
                        raise EasyBuildError("When using ensurepip to install pip (requested by install_pip=True) "
                                             "you must set 'use_pip=True' for the pip & setuptools extensions. "
                                             "Found 'use_pip=False' (maybe by default) for %s.",
                                             ext['name'])

    def auto_detect_lto_support(self):
        """Return True, if LTO should be enabled for current toolchain"""
        result = False
        # GCC >= 8 should be stable enough for LTO
        if self.toolchain.comp_family() == toolchain.GCC:
            gcc_ver = get_software_version('GCCcore') or get_software_version('GCC')
            if gcc_ver and LooseVersion(gcc_ver) >= LooseVersion('8.0'):
                self.log.info("Auto-enabling LTO since GCC >= v8.0 is used as toolchain compiler")
                result = True
        return result

    def _has_ensure_pip(self):
        """Check if  this Python version has/should have the ensurepip package"""
        # Pip is included since 3.4 via ensurepip https://docs.python.org/3.4/whatsnew/changelog.html
        # And in 2.7.9+: https://docs.python.org/2.7/whatsnew/2.7.html#pep-477-backport-ensurepip-pep-453-to-python-2-7
        version = LooseVersion(self.version)
        return version >= LooseVersion('3.4.0') or (version < LooseVersion('3') and version >= LooseVersion('2.7.9'))

    def configure_step(self):
        """Set extra configure options."""
        # Check for and report distutils user configs which may make the installation fail
        # See https://github.com/easybuilders/easybuild-easyconfigs/issues/11009
        for cfg in [os.path.join(os.path.expanduser('~'), name) for name in ('.pydistutils.cfg', 'pydistutils.cfg')]:
            if os.path.exists(cfg):
                raise EasyBuildError("Legacy distutils user configuration file found at %s. Aborting.", cfg)

        self.cfg.update('configopts', "--enable-shared")

        # Explicitely enable thread support on < 3.7 (always on 3.7+)
        if LooseVersion(self.version) < LooseVersion('3.7'):
            self.cfg.update('configopts', "--with-threads")

        # Explicitely enable unicode on Python 2, always on for Python 3
        # Need to be careful to match the unicode settings to the underlying python
        if LooseVersion(self.version) < LooseVersion('3.0'):
            if sys.maxunicode == 1114111:
                self.cfg.update('configopts', "--enable-unicode=ucs4")
            elif sys.maxunicode == 65535:
                self.cfg.update('configopts', "--enable-unicode=ucs2")
            else:
                raise EasyBuildError("Unknown maxunicode value for your python: %d" % sys.maxunicode)

        # LTO introduced in 3.7.0
        if LooseVersion(self.version) >= LooseVersion('3.7.0'):
            use_lto = self.cfg['use_lto']
            if use_lto is None:
                use_lto = self.auto_detect_lto_support()
            if use_lto:
                self.cfg.update('configopts', "--with-lto")

        # Enable further optimizations at the cost of a longer build
        # Introduced in 3.5.3, fixed in 3.5.4: https://docs.python.org/3.5/whatsnew/changelog.html
        if self.cfg['optimized'] and LooseVersion(self.version) >= LooseVersion('3.5.4'):
            # only configure with --enable-optimizations when compiling Python with (recent) GCC compiler
            if self.toolchain.comp_family() == toolchain.GCC:
                gcc_ver = get_software_version('GCCcore') or get_software_version('GCC')
                if LooseVersion(gcc_ver) >= LooseVersion('8.0'):
                    self.cfg.update('configopts', "--enable-optimizations")

        # When ensurepip is available we explicitely set this.
        # E.g. in 3.4 it is by default "upgrade", i.e. on which is unexpected when we did set it to off
        if self._has_ensure_pip():
            self.cfg.update('configopts', "--with-ensurepip=" + ('no', 'upgrade')[self.install_pip])

        modules_setup = os.path.join(self.cfg['start_dir'], 'Modules', 'Setup')
        if LooseVersion(self.version) < LooseVersion('3.8.0'):
            modules_setup += '.dist'

        libreadline = get_software_root('libreadline')
        if libreadline:
            ncurses = get_software_root('ncurses')
            if ncurses:
                readline_libdir = get_software_libdir('libreadline')
                ncurses_libdir = get_software_libdir('ncurses')
                readline_static_lib = os.path.join(libreadline, readline_libdir, 'libreadline.a')
                ncurses_static_lib = os.path.join(ncurses, ncurses_libdir, 'libncurses.a')
                readline = "readline readline.c %s %s" % (readline_static_lib, ncurses_static_lib)
                for line in fileinput.input(modules_setup, inplace='1', backup='.readline'):
                    line = re.sub(r"^#readline readline.c.*", readline, line)
                    sys.stdout.write(line)
            else:
                raise EasyBuildError("Both libreadline and ncurses are required to ensure readline support")

        openssl = get_software_root('OpenSSL')
        if openssl:
            for line in fileinput.input(modules_setup, inplace='1', backup='.ssl'):
                line = re.sub(r"^#SSL=.*", "SSL=%s" % openssl, line)
                line = re.sub(r"^#(\s*-DUSE_SSL -I)", r"\1", line)
                line = re.sub(r"^#(\s*-L\$\(SSL\)/lib )", r"\1 -L$(SSL)/lib64 ", line)
                sys.stdout.write(line)

        tcl = get_software_root('Tcl')
        tk = get_software_root('Tk')
        if tcl and tk:
            tclver = get_software_version('Tcl')
            tkver = get_software_version('Tk')
            tcltk_maj_min_ver = '.'.join(tclver.split('.')[:2])
            if tcltk_maj_min_ver != '.'.join(tkver.split('.')[:2]):
                raise EasyBuildError("Tcl and Tk major/minor versions don't match: %s vs %s", tclver, tkver)

            self.cfg.update('configopts', "--with-tcltk-includes='-I%s/include -I%s/include'" % (tcl, tk))

            tcl_libdir = os.path.join(tcl, get_software_libdir('Tcl'))
            tk_libdir = os.path.join(tk, get_software_libdir('Tk'))
            tcltk_libs = "-L%(tcl_libdir)s -L%(tk_libdir)s -ltcl%(maj_min_ver)s -ltk%(maj_min_ver)s" % {
                'tcl_libdir': tcl_libdir,
                'tk_libdir': tk_libdir,
                'maj_min_ver': tcltk_maj_min_ver,
            }
            self.cfg.update('configopts', "--with-tcltk-libs='%s'" % tcltk_libs)

        # don't add user site directory to sys.path (equivalent to python -s)
        # This matters e.g. when python installs the bundled pip & setuptools (for >= 3.4)
        env.setvar('PYTHONNOUSERSITE', '1')

        super(EB_Python, self).configure_step()

    def build_step(self, *args, **kwargs):
        """Custom build procedure for Python, ensure stack size limit is set to 'unlimited' (if desired)."""

        # make sure installation directory doesn't already exist when building with --rpath and
        # configuring with --enable-optimizations, since that leads to errors like:
        #   ./python: symbol lookup error: ./python: undefined symbol: __gcov_indirect_call
        # see also https://bugs.python.org/issue29712
        enable_opts_flag = '--enable-optimizations'
        if build_option('rpath') and enable_opts_flag in self.cfg['configopts']:
            if os.path.exists(self.installdir):
                warning_msg = "Removing existing installation directory '%s', "
                warning_msg += "because EasyBuild is configured to use RPATH linking "
                warning_msg += "and %s configure option is used." % enable_opts_flag
                print_warning(warning_msg % self.installdir)
                remove_dir(self.installdir)

        if self.cfg['ulimit_unlimited']:
            # determine current stack size limit
            (out, _) = run_cmd("ulimit -s")
            curr_ulimit_s = out.strip()

            # figure out hard limit for stack size limit;
            # this determines whether or not we can use "ulimit -s unlimited"
            (out, _) = run_cmd("ulimit -s -H")
            max_ulimit_s = out.strip()

            if curr_ulimit_s == UNLIMITED:
                self.log.info("Current stack size limit is %s: OK", curr_ulimit_s)
            elif max_ulimit_s == UNLIMITED:
                self.log.info("Current stack size limit is %s, setting it to %s for build...",
                              curr_ulimit_s, UNLIMITED)
                self.cfg.update('prebuildopts', "ulimit -s %s && " % UNLIMITED)
            else:
                msg = "Current stack size limit is %s, and can not be set to %s due to hard limit of %s;"
                msg += " setting stack size limit to %s instead, "
                msg += " this may break part of the compilation (e.g. hashlib)..."
                print_warning(msg % (curr_ulimit_s, UNLIMITED, max_ulimit_s, max_ulimit_s))
                self.cfg.update('prebuildopts', "ulimit -s %s && " % max_ulimit_s)

        super(EB_Python, self).build_step(*args, **kwargs)

    def install_step(self):
        """Extend make install to make sure that the 'python' command is present."""

        # avoid that pip (ab)uses $HOME/.cache/pip
        # cfr. https://pip.pypa.io/en/stable/reference/pip_install/#caching
        env.setvar('XDG_CACHE_HOME', tempfile.gettempdir())
        self.log.info("Using %s as pip cache directory", os.environ['XDG_CACHE_HOME'])

        super(EB_Python, self).install_step()

        # Create non-versioned, relative symlinks for python, python-config and pip
        python_binary_path = os.path.join(self.installdir, 'bin', 'python')
        if not os.path.isfile(python_binary_path):
            symlink('python' + self.pyshortver, python_binary_path, use_abspath_source=False)
        python_config_binary_path = os.path.join(self.installdir, 'bin', 'python-config')
        if not os.path.isfile(python_config_binary_path):
            symlink('python' + self.pyshortver + '-config', python_config_binary_path, use_abspath_source=False)
        if self.install_pip:
            pip_binary_path = os.path.join(self.installdir, 'bin', 'pip')
            if not os.path.isfile(pip_binary_path):
                symlink('pip' + self.pyshortver, pip_binary_path, use_abspath_source=False)

        if self.cfg.get('ebpythonprefixes'):
            write_file(os.path.join(self.installdir, self.pythonpath, 'sitecustomize.py'), SITECUSTOMIZE)

        # symlink lib/python*/lib-dynload to lib64/python*/lib-dynload if it doesn't exist;
        # see https://github.com/easybuilders/easybuild-easyblocks/issues/1957
        lib_dynload = 'lib-dynload'
        python_lib_dynload = os.path.join('python%s' % self.pyshortver, lib_dynload)
        lib_dynload_path = os.path.join(self.installdir, 'lib', python_lib_dynload)
        if not os.path.exists(lib_dynload_path):
            lib64_dynload_path = os.path.join('lib64', python_lib_dynload)
            if os.path.exists(os.path.join(self.installdir, lib64_dynload_path)):
                lib_dynload_parent = os.path.dirname(lib_dynload_path)
                mkdir(lib_dynload_parent, parents=True)
                cwd = change_dir(lib_dynload_parent)
                # use relative path as target, to avoid hardcoding path to install directory
                target_lib_dynload = os.path.join('..', '..', lib64_dynload_path)
                symlink(target_lib_dynload, lib_dynload)
                change_dir(cwd)

    def _sanity_check_ebpythonprefixes(self):
        """Check that EBPYTHONPREFIXES works"""
        temp_prefix = tempfile.mkdtemp(suffix='-tmp-prefix')
        site_packages_path = os.path.join('lib', 'python' + self.pyshortver, 'site-packages')
        temp_site_packages_path = os.path.join(temp_prefix, site_packages_path)
        mkdir(temp_site_packages_path, parents=True)  # Must exist
        (out, _) = run_cmd("%s=%s python -c 'import sys; print(sys.path)'" % (EBPYTHONPREFIXES, temp_prefix))
        out = out.strip()
        # Output should be a list which we can evaluate directly
        if not out.startswith('[') or not out.endswith(']'):
            raise EasyBuildError("Unexpected output for sys.path: %s", out)
        paths = eval(out)
        base_site_packages_path = os.path.join(self.installdir, site_packages_path)
        try:
            base_prefix_idx = paths.index(base_site_packages_path)
        except ValueError:
            raise EasyBuildError("The Python install path was not added to sys.path (%s)", paths)
        try:
            eb_prefix_idx = paths.index(temp_site_packages_path)
        except ValueError:
            raise EasyBuildError("EasyBuilds sitecustomize.py did not add %s to sys.path (%s)",
                                 temp_site_packages_path, paths)
        if eb_prefix_idx > base_prefix_idx:
            raise EasyBuildError("EasyBuilds sitecustomize.py did not add %s before %s to sys.path (%s)",
                                 temp_site_packages_path, base_site_packages_path, paths)

    def sanity_check_step(self):
        """Custom sanity check for Python."""

        shlib_ext = get_shared_lib_ext()

        try:
            fake_mod_data = self.load_fake_module()
        except EasyBuildError as err:
            raise EasyBuildError("Loading fake module failed: %s", err)

        abiflags = ''
        if LooseVersion(self.version) >= LooseVersion("3"):
            run_cmd("command -v python", log_all=True, simple=False, trace=False)
            cmd = 'python -c "import sysconfig; print(sysconfig.get_config_var(\'abiflags\'));"'
            (abiflags, _) = run_cmd(cmd, log_all=True, simple=False, trace=False)
            if not abiflags:
                raise EasyBuildError("Failed to determine abiflags: %s", abiflags)
            else:
                abiflags = abiflags.strip()

        # make sure hashlib is installed correctly, there should be no errors/output when 'import hashlib' is run
        # (python will exit with 0 regardless of whether or not errors are printed...)
        # cfr. https://github.com/easybuilders/easybuild-easyconfigs/issues/6484
        cmd = "python -c 'import hashlib'"
        (out, _) = run_cmd(cmd)
        regex = re.compile('error', re.I)
        if regex.search(out):
            raise EasyBuildError("Found one or more errors in output of %s: %s", cmd, out)
        else:
            self.log.info("No errors found in output of %s: %s", cmd, out)

        if self.cfg.get('ebpythonprefixes'):
            self._sanity_check_ebpythonprefixes()

        pyver = 'python' + self.pyshortver
        custom_paths = {
            'files': [
                os.path.join('bin', pyver),
                os.path.join('bin', 'python'),
                os.path.join('bin', pyver + '-config'),
                os.path.join('bin', 'python-config'),
                os.path.join('lib', 'lib' + pyver + abiflags + '.' + shlib_ext),
            ],
            'dirs': [os.path.join('include', pyver + abiflags), os.path.join('lib', pyver, 'lib-dynload')],
        }

        # cleanup
        self.clean_up_fake_module(fake_mod_data)

        custom_commands = [
            "python --version",
            "python-config --help",  # make sure that symlink was created correctly
            "python -c 'import _ctypes'",  # make sure that foreign function interface (libffi) works
            "python -c 'import _ssl'",  # make sure SSL support is enabled one way or another
            "python -c 'import readline'",  # make sure readline support was built correctly
        ]

        if self.install_pip:
            # Check that pip and setuptools are installed
            py_maj_version = self.version.split('.')[0]
            custom_paths['files'].extend([
                os.path.join('bin', pip) for pip in ('pip', 'pip' + py_maj_version, 'pip' + self.pyshortver)
            ])
            custom_commands.extend([
                "python -c 'import pip'",
                "python -c 'import setuptools'",
            ])

        if get_software_root('Tk'):
            # also check whether importing tkinter module works, name is different for Python v2.x and v3.x
            if LooseVersion(self.version) >= LooseVersion('3'):
                tkinter = 'tkinter'
            else:
                tkinter = 'Tkinter'
            custom_commands.append("python -c 'import %s'" % tkinter)

            # check whether _tkinter*.so is found, exact filename doesn't matter
            tkinter_so = os.path.join(self.installdir, 'lib', pyver, 'lib-dynload', '_tkinter*.' + shlib_ext)
            tkinter_so_hits = glob.glob(tkinter_so)
            if len(tkinter_so_hits) == 1:
                self.log.info("Found exactly one _tkinter*.so: %s", tkinter_so_hits[0])
            else:
                raise EasyBuildError("Expected to find exactly one _tkinter*.so: %s", tkinter_so_hits)

        super(EB_Python, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_extra(self, *args, **kwargs):
        """Add path to sitecustomize.py to $PYTHONPATH"""
        txt = super(EB_Python, self).make_module_extra()

        if self.cfg.get('ebpythonprefixes'):
            txt += self.module_generator.prepend_paths('PYTHONPATH', self.pythonpath)

        return txt



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing NEURON, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Maxime Boissonneault (Universite Laval, Compute Canada)
"""
import os
import re

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.easyblocks.generic.cmakemake import CMakeMake
from easybuild.easyblocks.generic.pythonpackage import det_pylibdir
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.config import build_option
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext

from distutils.version import LooseVersion


class EB_NEURON(CMakeMake):
    """Support for building/installing NEURON."""

    def __init__(self, *args, **kwargs):
        """Initialisation of custom class variables for NEURON."""
        super(EB_NEURON, self).__init__(*args, **kwargs)

        self.hostcpu = 'UNKNOWN'
        self.with_python = False
        self.pylibdir = 'UNKNOWN'

    @staticmethod
    def extra_options():
        """Custom easyconfig parameters for NEURON."""

        extra_vars = {
            'paranrn': [True, "Enable support for distributed simulations.", CUSTOM],
        }
        return CMakeMake.extra_options(extra_vars)

    def configure_step(self):
        """Custom configuration procedure for NEURON."""
        if LooseVersion(self.version) < LooseVersion('7.8.1'):

            # make sure we're using the correct configure command
            # (required because custom easyconfig parameters from CMakeMake are picked up)
            self.cfg['configure_cmd'] = "./configure"

            # enable support for distributed simulations if desired
            if self.cfg['paranrn']:
                self.cfg.update('configopts', '--with-paranrn')

            # specify path to InterViews if it is available as a dependency
            interviews_root = get_software_root('InterViews')
            if interviews_root:
                self.cfg.update('configopts', "--with-iv=%s" % interviews_root)
            else:
                self.cfg.update('configopts', "--without-iv")

            # optionally enable support for Python as alternative interpreter
            python_root = get_software_root('Python')
            if python_root:
                self.with_python = True
                self.cfg.update('configopts', "--with-nrnpython=%s/bin/python" % python_root)

            # determine host CPU type
            cmd = "./config.guess"
            (out, ec) = run_cmd(cmd, simple=False)

            self.hostcpu = out.split('\n')[0].split('-')[0]
            self.log.debug("Determined host CPU type as %s" % self.hostcpu)

            # determine Python lib dir
            self.pylibdir = det_pylibdir()

            # complete configuration with configure_method of parent
            ConfigureMake.configure_step(self)
        else:
            # enable support for distributed simulations if desired
            if self.cfg['paranrn']:
                self.cfg.update('configopts', '-DNRN_ENABLE_MPI=ON')
            else:
                self.cfg.update('configopts', '-DNRN_ENABLE_MPI=OFF')

            # specify path to InterViews if it is available as a dependency
            interviews_root = get_software_root('InterViews')
            if interviews_root:
                self.cfg.update('configopts', "-DIV_DIR=%s -DNRN_ENABLE_INTERVIEWS=ON" % interviews_root)
            else:
                self.cfg.update('configopts', "-DNRN_ENABLE_INTERVIEWS=OFF")

            # no longer used it seems
            self.hostcpu = ''

            # optionally enable support for Python as alternative interpreter
            python_root = get_software_root('Python')
            if python_root:
                self.with_python = True
                self.cfg.update('configopts', "-DNRN_ENABLE_PYTHON=ON -DPYTHON_EXECUTABLE=%s/bin/python" % python_root)
                self.cfg.update('configopts', "-DNRN_ENABLE_MODULE_INSTALL=ON "
                                "-DNRN_MODULE_INSTALL_OPTIONS='--prefix=%s'" % self.installdir)
            else:
                self.cfg.update('configopts', "-DNRN_ENABLE_PYTHON=OFF")

            # determine Python lib dir
            self.pylibdir = det_pylibdir()

            # complete configuration with configure_method of parent
            CMakeMake.configure_step(self)

    def install_step(self):
        """Custom install procedure for NEURON."""

        super(EB_NEURON, self).install_step()

        # with the CMakeMake, the python module is installed automatically
        if LooseVersion(self.version) < LooseVersion('7.8.1'):
            if self.with_python:
                pypath = os.path.join('src', 'nrnpython')
                try:
                    pwd = os.getcwd()
                    os.chdir(pypath)
                except OSError as err:
                    raise EasyBuildError("Failed to change to %s: %s", pypath, err)

                cmd = "python setup.py install --prefix=%s" % self.installdir
                run_cmd(cmd, simple=True, log_all=True, log_ok=True)

                try:
                    os.chdir(pwd)
                except OSError as err:
                    raise EasyBuildError("Failed to change back to %s: %s", pwd, err)

    def sanity_check_step(self):
        """Custom sanity check for NEURON."""
        shlib_ext = get_shared_lib_ext()
        binpath = os.path.join(self.hostcpu, 'bin')
        libpath = os.path.join(self.hostcpu, 'lib', 'lib%s.' + shlib_ext)
        # hoc_ed is not included in the sources of 7.4. However, it is included in the binary distribution.
        # Nevertheless, the binary has a date old enough (June 2014, instead of November 2015 like all the
        # others) to be considered a mistake in the distribution
        binaries = ["neurondemo", "nrngui", "nrniv", "nrnivmodl", "nocmodl", "modlunit", "nrnmech_makefile",
                    "mkthreadsafe"]
        libs = ["nrniv"]
        sanity_check_dirs = ['share/nrn']

        if LooseVersion(self.version) < LooseVersion('7.4'):
            binaries += ["hoc_ed"]

        if LooseVersion(self.version) < LooseVersion('7.8.1'):
            binaries += ["bbswork.sh", "hel2mos1.sh", "ivoc", "memacs", "mos2nrn", "mos2nrn2.sh", "oc"]
            binaries += ["nrn%s" % x for x in ["iv_makefile", "oc", "oc_makefile", "ocmodl"]]
            libs += ["ivoc", "ivos", "memacs", "meschach", "neuron_gnu", "nrnmpi", "nrnoc", "nrnpython",
                     "oc", "ocxt", "scopmath", "sparse13", "sundials"]
            sanity_check_dirs += ['include/nrn']
        # list of included binaries changed with cmake. See
        # https://github.com/neuronsimulator/nrn/issues/899
        else:
            binaries += ["nrnpyenv.sh", "set_nrnpyenv.sh", "sortspike"]
            libs += ["rxdmath"]
            sanity_check_dirs += ['include']
            if self.with_python:
                sanity_check_dirs += [os.path.join("lib", "python"),
                                      os.path.join("lib", "python%(pyshortver)s", "site-packages")]

        # this is relevant for installations of Python packages for multiple Python versions (via multi_deps)
        # (we can not pass this via custom_paths, since then the %(pyshortver)s template value will not be resolved)
        # ensure that we only add to paths specified in the EasyConfig
        sanity_check_files = [os.path.join(binpath, x) for x in binaries] + [libpath % x for x in libs]
        self.cfg['sanity_check_paths'] = {
                'files': sanity_check_files,
                'dirs': sanity_check_dirs,
        }

        super(EB_NEURON, self).sanity_check_step()

        try:
            fake_mod_data = self.load_fake_module()
        except EasyBuildError as err:
            self.log.debug("Loading fake module failed: %s" % err)

        # test NEURON demo
        inp = '\n'.join([
            "demo(3) // load the pyramidal cell model.",
            "init()  // initialise the model",
            "t       // should be zero",
            "soma.v  // will print -65",
            "run()   // run the simulation",
            "t       // should be 5, indicating that 5ms were simulated",
            "soma.v  // will print a value other than -65, indicating that the simulation was executed",
            "quit()",
        ])
        (out, ec) = run_cmd("neurondemo", simple=False, log_all=True, log_output=True, inp=inp)

        validate_regexp = re.compile(r"^\s+-65\s*\n\s+5\s*\n\s+-68.134337", re.M)
        if ec or not validate_regexp.search(out):
            raise EasyBuildError("Validation of NEURON demo run failed.")
        else:
            self.log.info("Validation of NEURON demo OK!")

        if build_option('mpi_tests'):
            nproc = self.cfg['parallel']
            try:
                cwd = os.getcwd()
                os.chdir(os.path.join(self.cfg['start_dir'], 'src', 'parallel'))

                cmd = self.toolchain.mpi_cmd_for("nrniv -mpi test0.hoc", nproc)
                (out, ec) = run_cmd(cmd, simple=False, log_all=True, log_output=True)

                os.chdir(cwd)
            except OSError as err:
                raise EasyBuildError("Failed to run parallel hello world: %s", err)

            valid = True
            for i in range(0, nproc):
                validate_regexp = re.compile("I am %d of %d" % (i, nproc))
                if not validate_regexp.search(out):
                    valid = False
                    break
            if ec or not valid:
                raise EasyBuildError("Validation of parallel hello world run failed.")
            else:
                self.log.info("Parallel hello world OK!")
        else:
            self.log.info("Skipping MPI testing of NEURON since MPI testing is disabled")

        if self.with_python:
            cmd = "python -c 'import neuron; neuron.test()'"
            (out, ec) = run_cmd(cmd, simple=False, log_all=True, log_output=True)

        # cleanup
        self.clean_up_fake_module(fake_mod_data)

    def make_module_req_guess(self):
        """Custom guesses for environment variables (PATH, ...) for NEURON."""

        guesses = super(EB_NEURON, self).make_module_req_guess()

        guesses.update({
            'PATH': [os.path.join(self.hostcpu, 'bin')],
        })

        return guesses

    def make_module_extra(self):
        """Define extra module entries required."""

        txt = super(EB_NEURON, self).make_module_extra()

        # we need to make sure the correct compiler is set in the environment,
        # since NEURON features compilation at runtime
        for var in ['CC', 'CXX', 'MPICC', 'MPICXX', 'MPICH_CC', 'MPICH_CXX']:
            val = os.getenv(var)
            if val:
                txt += self.module_generator.set_environment(var, val)
                self.log.debug("%s set to %s, adding it to module" % (var, val))
            else:
                self.log.debug("%s not set: %s" % (var, os.environ.get(var, None)))

        if self.with_python:
            if self.cfg['multi_deps'] and 'Python' in self.cfg['multi_deps']:
                txt += self.module_generator.prepend_paths('EBPYTHONPREFIXES', '')
            else:
                txt += self.module_generator.prepend_paths('PYTHONPATH', [self.pylibdir])
            # also adds lib/python to PYTHONPATH
            txt += self.module_generator.prepend_paths('PYTHONPATH', ['lib/python'])
        return txt



# #
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #
"""
EasyBuild support for installing the Intel Trace Analyzer and Collector (ITAC), implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
"""

import os

from distutils.version import LooseVersion

from easybuild.easyblocks.generic.intelbase import IntelBase
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.run import run_cmd


class EB_itac(IntelBase):
    """
    Class that can be used to install itac
    - tested with Intel Trace Analyzer and Collector 7.2.1.008
    """

    @staticmethod
    def extra_options():
        extra_vars = {
            'preferredmpi': ['impi3', "Preferred MPI type", CUSTOM],
        }
        return IntelBase.extra_options(extra_vars)

    def prepare_step(self, *args, **kwargs):
        """
        Custom prepare step for itac: don't require runtime license for oneAPI versions (>= 2021)
        """
        if LooseVersion(self.version) >= LooseVersion('2021'):
            kwargs['requires_runtime_license'] = False

        super(EB_itac, self).prepare_step(*args, **kwargs)

    def install_step_classic(self):
        """
        Actual installation for versions prior to 2021.x

        - create silent cfg file
        - execute command
        """

        if LooseVersion(self.version) >= LooseVersion('8.1'):
            super(EB_itac, self).install_step_classic(silent_cfg_names_map=None)

            # itac v9.0.1 installer create itac/<version> subdir, so stuff needs to be moved afterwards
            if LooseVersion(self.version) >= LooseVersion('9.0'):
                super(EB_itac, self).move_after_install()
        else:
            silent = """
[itac]
INSTALLDIR=%(ins)s
LICENSEPATH=%(lic)s
INSTALLMODE=NONRPM
INSTALLUSER=NONROOT
INSTALL_ITA=YES
INSTALL_ITC=YES
DEFAULT_MPI=%(mpi)s
EULA=accept
""" % {'lic': self.license_file, 'ins': self.installdir, 'mpi': self.cfg['preferredmpi']}

            # already in correct directory
            silentcfg = os.path.join(os.getcwd(), "silent.cfg")
            f = open(silentcfg, 'w')
            f.write(silent)
            f.close()
            self.log.debug("Contents of %s: %s" % (silentcfg, silent))

            tmpdir = os.path.join(os.getcwd(), self.version, 'mytmpdir')
            try:
                os.makedirs(tmpdir)
            except OSError as err:
                raise EasyBuildError("Directory %s can't be created: %s", tmpdir, err)

            cmd = "./install.sh --tmp-dir=%s --silent=%s" % (tmpdir, silentcfg)

            run_cmd(cmd, log_all=True, simple=True)

    def install_step_oneapi(self, *args, **kwargs):
        """
        Actual installation for versions 2021.x onwards.
        """
        # require that EULA is accepted
        intel_eula_url = 'https://software.intel.com/content/www/us/en/develop/articles/end-user-license-agreement.html'
        self.check_accepted_eula(name='Intel-oneAPI', more_info=intel_eula_url)

        # exactly one "source" file is expected: the (offline) installation script
        if len(self.src) == 1:
            install_script = self.src[0]['name']
        else:
            src_fns = ', '.join([x['name'] for x in self.src])
            raise EasyBuildError("Expected to find exactly one 'source' file (installation script): %s", src_fns)

        cmd = ' '.join([
            "sh %s" % install_script,
            '-a',
            '-s',
            "--eula accept",
            "--install-dir=%s" % self.installdir,
        ])

        run_cmd(cmd, log_all=True, simple=True)

        # itac installer create itac/<version> subdir, so stuff needs to be moved afterwards
        super(EB_itac, self).move_after_install()

    def sanity_check_step(self):
        """Custom sanity check paths for ITAC."""

        custom_paths = {
            'files': ["include/%s" % x for x in ["i_malloc.h", "VT_dynamic.h", "VT.h", "VT.inc"]],
            'dirs': ["bin", "lib", "slib"],
        }

        super(EB_itac, self).sanity_check_step(custom_paths=custom_paths)

    def make_module_req_guess(self):
        """
        A dictionary of possible directories to look for
        """
        guesses = {}
        if LooseVersion(self.version) < LooseVersion('9.0'):
            preferredmpi = self.cfg["preferredmpi"]
            guesses.update({
                'MANPATH': ['man'],
                'CLASSPATH': ['itac/lib_%s' % preferredmpi],
                'VT_LIB_DIR': ['itac/lib_%s' % preferredmpi],
                'VT_SLIB_DIR': ['itac/lib_s%s' % preferredmpi]
            })

        if self.cfg['m32']:
            guesses.update({
                'PATH': ['bin', 'bin/ia32', 'ia32/bin'],
                'LD_LIBRARY_PATH': ['lib', 'lib/ia32', 'ia32/lib'],
            })
        else:
            guesses.update({
                'PATH': ['bin', 'bin/intel64', 'bin64'],
                'LD_LIBRARY_PATH': ['lib', 'lib/intel64', 'lib64', 'slib'],
            })
        return guesses

    def make_module_extra(self):
        """Overwritten from IntelBase to add extra txt"""
        txt = super(EB_itac, self).make_module_extra()
        txt += self.module_generator.set_environment('VT_ROOT', self.installdir)
        txt += self.module_generator.set_environment('VT_MPI', self.cfg['preferredmpi'])
        txt += self.module_generator.set_environment('VT_ADD_LIBS', "-ldwarf -lelf -lvtunwind -lnsl -lm -ldl -lpthread")
        txt += self.module_generator.set_environment('VT_LIB_DIR', self.installdir + "/lib")
        txt += self.module_generator.set_environment('VT_SLIB_DIR', self.installdir + "/slib")
        return txt
